-module(stm).
-export([start/1, stop/1, 
         create/3, create/2, drop/2, 
         pull/2, push/2, 
         read/1, write/2,
         cstart/1, cpull/2, cpush/2, open/1, close/1, atomic/3,
         qnew/2, qpop/2, qpush/3, qtransfer/3,
         qtest/0]).


% Variable: {name, version, value}
-record(var, {name, version = 0, value = undefined}).
% TM server state: storage and open transactions
-record(state, {storage, trans}).


%%%
%%% Non-compositional transactional memory
%%% 

% initialize empty transactional memory and register as `Tm'
start(Tm) -> 
  Storage = dict:new(),  % empty key/value map
  register(Tm,
    gserver:start(#state{storage = Storage},
                  fun stm/2)).

% shutdown `Tm'
stop(Tm) ->
  gserver:stop(Tm).

% create variable `Name' in `Tm'
% using initial `Value'
% return:
%  ok -> success
%  skip -> variable `Name' already exists
create(Tm, Name, Value) ->
  gserver:request(Tm, {create, Name, Value}).

% create variable `Name' in `Tm'
% with `undefined' initial value
% return:
%  ok -> success
%  skip -> variable `Name' already exists
create(Tm, Name) ->
  create(Tm, Name, undefined).

% remove variable `Name' from `Tm'
% return:
%  ok -> success
%  skip -> variable `Name' already exists
drop(Tm, Name) ->
  gserver:request(Tm, {drop, Name}).

% check out variable `Name' from `Tm'
% return:
%  variable -> variable `Name' found
%  not_found -> variable `Name' not found
pull(Tm, Name) ->
  gserver:request(Tm, {pull, Name}).

% commit all variables in list `Vars' to `Tm'
% return:
%  success -> changes committed
%  fail -> changes not committed, abort
push(Tm, Vars) when is_list(Vars) ->
  gserver:request(Tm, {push, Vars});
% commit variable Var to `Tm'
push(Tm, Var) ->
  push(Tm, [Var]).

% read value of variable
read(#var{value = Value}) ->
  Value.

% write `Value' to `Var'
write(Var = #var{}, Value) ->
  Var#var{value = Value}.


stm(State = #state{storage = Storage},
    {create, Name, Value}) ->
  case dict:is_key(Name, Storage) of
	 true ->
      % variable `Name' already exists
      {reply, State, skip};
	 false ->
      % add variable
      Var = #var{name = Name, value = Value},
      {reply, State#state{storage = 
              dict:store(Name, Var, Storage)}, ok}
  end;

stm(State = #state{storage = Storage},
    {drop, Name}) ->
  case dict:is_key(Name, Storage) of
	 true ->
      {reply, State#state{storage = 
              dict:erase(Name, Storage)}, ok};
	 false ->
      % variable `Name' does not exist
      {reply, State, skip}
  end;

stm(State = #state{storage = Storage},
    {pull, Name}) ->
  case dict:is_key(Name, Storage) of
    true ->
      {reply, State, dict:fetch(Name, Storage)};
    false ->
      {reply, State, not_found}
   end;

stm(State = #state{storage = Storage},
    {push, Vars}) ->
  case try_push(Vars, Storage) of
    {success, NewStorage} ->
      {reply, State#state{storage = 
              NewStorage}, success};
	 fail ->
      {reply, State, fail}
  end.

try_push([], Storage) ->
  {success, Storage};
try_push([Var = #var{name = Name, version = Version} | Vars], 
          Storage) ->
  case dict:find(Name, Storage) of
    {ok, #var{version = Version}} ->
      try_push(Vars,
               dict:store(Name, Var#var{version = Version + 1}, Storage));
	 _ -> fail
  end.


%%%
%%% Compositional transactional memory
%%% 

cstart(Tm) -> 
  Storage = dict:new(),  % empty key/value map
  Trans = dict:new(),
  register(Tm,
    gserver:start(#state{storage = Storage, trans = Trans},
                  fun cstm/2)).

% composable pull
% return:
%  variable -> variable `Name' found
%  not_found -> variable `Name' not found
cpull(Tm, Name) ->
  Me = self(),
  gserver:request(Tm, {pull, Name, Me}).

% composable push
% return:
%  success -> changes committed
%  fail -> changes not committed, abort
%  defer -> changes deferred to transaction end
cpush(Tm, Vars) when is_list(Vars) ->
  Me = self(),
	 gserver:request(Tm, {push, Vars, Me});
cpush(Tm, Var) ->
  cpush(Tm, [Var]).

% begin composite transaction
% return:
%  ok -> transaction opened
open(Tm) ->
  Me = self(),
  gserver:request(Tm, {open, Me}).

% end composite transaction
% return:
%  success -> changes committed, transaction closed
%  fail -> changes not committed, trasaction aborted
%  pending -> nested transactions pending, current level closed
%  skip -> transaction never opened: do nothing
close(Tm) ->
  Me = self(),
  gserver:request(Tm, {close, Me}).


% atomically execute Function on arguments Args 
% (the code must use cpull and cpush, and Tm must be a cstm)
% atomic transactions can be nested
atomic(Tm, Function, Args) ->
  open(Tm),
  apply(Function, Args),
  case close(Tm) of
    fail -> atomic(Tm, Function, Args);
	 _ -> ok
  end.


%% trans is a dictionary Pid -> Transaction
%% a transaction is a dictionary VarNames -> Vars

cstm(State = #state{trans = Trans},
     {open, Who}) ->
  case dict:find(Who, Trans) of
    % a transaction by Who is already pending
    {ok, {N, Vars}} ->
      % increment the nesting level
      NewTr = {N + 1, Vars};
    % Who has no pending transactions
	 _ ->
      % create a new transaction by Who
      NewTr = {1, dict:new()}
  end,
  % add or replace transaction
  NewTrans = dict:store(Who, NewTr, Trans),
  % open always succeeds
  {reply, State#state{trans = NewTrans}, ok};	 

cstm(State = #state{trans = Trans},
     {close, Who}) ->
  case dict:find(Who, Trans) of
    % Who is closing its last transaction level
    {ok, {N, Tr}} when N =:= 1 -> % commit all changes
        % remove transaction
      NewState = State#state{trans = dict:erase(Who, Trans)},
        % variables to be committed
      Vars = [Val || {_, Val} <- dict:to_list(Tr)],
        % try to push: it may fail
      stm(NewState, {push, Vars});
    % Who is closing its last transaction level
    {ok, {N, Tr}} -> % decrease transaction level
      NewTr = {N - 1, Tr},
       % update Who's transaction
      {reply, State#state{trans = 
         dict:store(Who, NewTr, Trans)}, pending};
    % Who is closing a transaction it never opened
	 _ -> % do nothing
      {reply, State, skip}
  end;

cstm(State = #state{storage = Storage, trans = Trans},
     {pull, Name, Who}) ->
  case dict:find(Who, Trans) of
    % there is an open transaction by Who
    {ok, {_N, Tr}} ->
      % pick the most recent value known to Who
      case latest(Name, Storage, Tr) of
          % a variable Name doesn't exist
        fail -> {reply, State, not_found};
          % variable found
        Var  -> {reply, State, Var}
      end;
    % there are no open transactions by Who
    _ -> % normal check out
      stm(State, {pull, Name})
  end;

cstm(State = #state{trans = Trans},
     {push, Vars, Who}) ->
  case dict:find(Who, Trans) of
    % there is an open transaction by Who
    {ok, {N, Tr}} -> % do not commit now
      % store variables in Who's transaction
      NewTr = {N, defer_push(Vars, Tr)},
      {reply, State#state{trans = 
        dict:store(Who, NewTr, Trans)}, defer};
    % there are no open transactions by Who
    _ -> % normal commit
      stm(State, {push, Vars})
  end;

% any other command on a cstm behaves as on a stm
cstm(State = #state{}, Command) ->
  stm(State, Command).


% find latest variable `Name' in `Tr' or `Storage'
latest(Name, Storage, Tr) ->
  case {dict:find(Name, Tr),
        dict:find(Name, Storage)} of
    {{ok, Var}, _} -> Var;  % if it's in Tr return it
    {_, {ok, Var}} -> Var;  % otherwise it should be in Storage
    _              -> fail  % a variable Name could not be found
  end.

% update `Tr' to store list of variables
% if a previous version of the variables exist, replace it
% so that the transaction will remain atomic
defer_push([], Tr) -> Tr;
defer_push([Var = #var{name = Name}|Vars], Tr) -> 
  defer_push(Vars, dict:store(Name, Var, Tr)).


%%%
%%% Using transactional memory: a concurrent queue
%%% 

% create new queue `Name'
qnew(Tm, Name) ->
  create(Tm, Name, []).

% pop head element from queue `Name'
qpop(Tm, Name) ->
  Queue = cpull(Tm, Name),
  [H|T] = read(Queue),
  NewQueue = write(Queue, T),
  case cpush(Tm, NewQueue) of
     % push failed: retry!
    fail -> qpop(Tm, Name);
     % push successful: return head
	 _ -> H 
  end.

% push `Value' to back of queue `Name'
qpush(Tm, Name, Value) ->
  Queue = cpull(Tm, Name),
  Vals = read(Queue),
  NewQueue = write(Queue, Vals ++ [Value]),
  case cpush(Tm, NewQueue) of
     % push failed: retry!
     fail -> qpush(Tm, Name, Value);
     % push successful: return ok
	  _ -> ok
  end.

% pop from Queue1 and push to Queue2
qtransfer_(Tm, Queue1, Queue2) ->
  Value = qpop(Tm, Queue1),
  qpush(Tm, Queue2, Value).

% atomic transfer
qtransfer(Tm, Queue1, Queue2) ->
	 atomic(Tm, fun qtransfer_/3, [Tm, Queue1, Queue2]).


% test atomic transactions
qtest() ->
  Tm = stm_,
  case whereis(Tm) of
    undefined -> skip;
	 _ ->
      stop(Tm),
      unregister(Tm)
  end,
  cstart(Tm),
  qnew(Tm, q1),
  qnew(Tm, q2),
  [qpush(Tm, q1, V) || V <- [1,2,3,4]],
  io:format("Popped: ~p (should be 1)~n", [qpop(Tm, q1)]),
  qtransfer(Tm, q1, q2),
  Q1 = read(cpull(Tm, q1)),
  Q2 = read(cpull(Tm, q2)),
  io:format("Transferred: q1 is ~p, q2 is ~p (should be [3,4], [2])~n", [Q1, Q2]),
  atomic(Tm, fun () -> qtransfer(Tm, q1, q2), qtransfer(Tm, q1, q2) end, []),
  NQ1 = read(cpull(Tm, q1)),
  NQ2 = read(cpull(Tm, q2)),
  io:format("Combined atomic: q1 is ~p, q2 is ~p (should be [], [2,3,4])~n", 
            [NQ1, NQ2]).
  			  

	 
	 
	 
	 
    
  
